/* ******************************************************************************
**
** Copyright (C) 1991-2008 Heinz van Saanen
**
** This file is part of the the function viewer Zhu3D.
**
** This file may be used under the terms of the GNU General Public
** License version 2 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
********************************************************************************** */

#include <QtGui>
#include "mainwindow.h"
#include "timers.h"			// CPU-specific timer code
#include "paths.h"			// Get doc/lang/app-paths	
#include "pinterface.h"		// Interface to parsers


// When 0.0f after main(), CPU counter is not implemented
// readSettings() gives a message during 1.st start then
extern float MHz;

// Be aware of current localzations
extern QString zhuLocale;

// Current default directories fetched from dirEdi
extern QString texDir;
extern QString workDir;

// Current state for internal/external browser from fntEdi
extern bool extBrowser;


/* ********************************************************************** */
/* MAINWINDOW CLASS                                                       */
/* ********************************************************************** */
MaiWindow::MaiWindow() {

	// Set up the main control-center
	controlCenter = new QWidget;
	Ui_mainCtrl::setupUi(controlCenter);
	this->setMinimumSize(250, 330);
	setCentralWidget(controlCenter);

	// Initialize the mode-button
	Parameter->setIcon(QPixmap(QString(":/images/player_fwd.png")));
	Parameter->setIconSize(QSize(18,18));

	// Create help/solver widgets
	hlpEdi = new QTextBrowser();
	solEdi = new QTextEdit();
	solEdi->setWindowTitle(tr("Solver report"));
	solEdi->setMinimumSize(270, 150);

	// Get stored settings like size, pos, ...
	readSettings();

	// Create actions. Done now, because of immediate file loading
	createActions();
	createMenus();
	createToolBars();

	// Done now, because of immediate file loading
	connect(this, SIGNAL(gotNewFile()), oglWid, SLOT(newFileOpened()));
	connect(this, SIGNAL(gotNewView()), oglWid, SLOT(updateGLonly()));

	// OpenGL now, because texture-size check may need it
	oglWid->show();
	sfineChecked=false;
	fineChecked=false;
	normalChecked=true;
	coarseChecked=false;

	// Load file argument or startup-file
	if(gotArg)
		if(loadFile(argStr))
			performOGLUpdate();
		else
			newSlot();
	else
		if(loadFile(MYSYSDIR+"startup.zhu"))
			performOGLUpdate();
		else
			newSlot();

	// More updates
	updMaiWid();			// Update MaiWindow-widget
	updMaiInf();			// Update MaiWindow-info-messages
	speEdi->updSpeWid();	// Texture/fog/motion blur stuff

	// CONNECTIONS
	// --------------------------------------------------------------------
	// Slider-signals and rotations-slots
	connect(xSlider, SIGNAL(valueChanged(int)), oglWid, SLOT(setXRotation(int)));
	connect(ySlider, SIGNAL(valueChanged(int)), oglWid, SLOT(setYRotation(int)));
	connect(zSlider, SIGNAL(valueChanged(int)), oglWid, SLOT(setZRotation(int)));

	// Correct slider-positions after mouse-signals and animation
	connect(oglWid, SIGNAL(xRotationChanged(int)), xSlider, SLOT(setValue(int)));
	connect(oglWid, SIGNAL(yRotationChanged(int)), ySlider, SLOT(setValue(int)));
	connect(aniEdi, SIGNAL(xRotationChanged(int)), xSlider, SLOT(setValue(int)));
	connect(aniEdi, SIGNAL(yRotationChanged(int)), ySlider, SLOT(setValue(int)));
	connect(aniEdi, SIGNAL(zRotationChanged(int)), zSlider, SLOT(setValue(int)));

	// Set modes
	connect(Point,     SIGNAL(clicked()), oglWid, SLOT(drawModePoint()));
	connect(Triangle,  SIGNAL(clicked()), oglWid, SLOT(drawModeTriangle()));
	connect(Quad,      SIGNAL(clicked()), oglWid, SLOT(drawModeQuad()));
	connect(Parameter, SIGNAL(clicked()), oglWid, SLOT(drawModeParameter()));
	connect(Point,     SIGNAL(clicked()), this, SLOT(wireModeSlot()));
	connect(Triangle,  SIGNAL(clicked()), this, SLOT(wireModeSlot()));
	connect(Quad,      SIGNAL(clicked()), this, SLOT(wireModeSlot()));
	connect(Parameter, SIGNAL(clicked()), this, SLOT(wireModeSlot()));

	// Enable/disable functions, cross ...
	connect(Function0, SIGNAL(clicked()), oglWid, SLOT(toggleF0()));	// Functions
	connect(Function1, SIGNAL(clicked()), oglWid, SLOT(toggleF1()));
	connect(Function2, SIGNAL(clicked()), oglWid, SLOT(toggleF2()));
	connect(Function3, SIGNAL(clicked()), oglWid, SLOT(toggleF3()));	// Parameters

	connect(Cross, SIGNAL(clicked()), oglWid, SLOT(toggleCross()));
	connect(Axes,  SIGNAL(clicked()), oglWid, SLOT(toggleAxes()));
	connect(Wired, SIGNAL(clicked()), oglWid, SLOT(togglePolygon()));

	// Handle scaling
	connect(xPlus,     SIGNAL(clicked()), oglWid, SLOT(scaleXplus()));
	connect(xMinus,    SIGNAL(clicked()), oglWid, SLOT(scaleXminus()));
	connect(yPlus,     SIGNAL(clicked()), oglWid, SLOT(scaleYplus()));
	connect(yMinus,    SIGNAL(clicked()), oglWid, SLOT(scaleYminus()));
	connect(zPlus,     SIGNAL(clicked()), oglWid, SLOT(scaleZplus()));
	connect(zMinus,    SIGNAL(clicked()), oglWid, SLOT(scaleZminus()));
	connect(xyzPlus,   SIGNAL(clicked()), oglWid, SLOT(scaleXYZplus()));
	connect(xyzMinus,  SIGNAL(clicked()), oglWid, SLOT(scaleXYZminus()));
	connect(gridPlus,  SIGNAL(clicked()), oglWid, SLOT(scaleGridplus()));
	connect(gridMinus, SIGNAL(clicked()), oglWid, SLOT(scaleGridminus()));

	// Close MaiWindow too, when OpenGL-viewer is closed
	connect(oglWid, SIGNAL(killAll()), this, SLOT(close()));

	// Some more commands from other widgets
	connect(oglWid, SIGNAL(updMaiWid()), this, SLOT(updMaiWid()));
	connect(oglWid, SIGNAL(updMaiInf()), this, SLOT(updMaiInf()));
	connect(aniEdi, SIGNAL(updMaiWid()), this, SLOT(updMaiWid()));
	connect(aniEdi, SIGNAL(updMaiInf()), this, SLOT(updMaiInf()));
	connect(aniEdi, SIGNAL(saveFromAni()), this, SLOT(saveAsSlot()));
	connect(fntEdi, SIGNAL(repaintMain()), this, SLOT(repaintSlot()));
	connect(fntEdi, SIGNAL(iconSize(const int)), this, SLOT(iconSizeSlot(const int)));

	// Parsers tainted, slide show running
	connect(usrEdi, SIGNAL(parsersTainted(bool)), this, SLOT(lockMainSlot(bool)));
	connect(oglWid, SIGNAL(parsersTainted(bool)), this, SLOT(lockMainSlot(bool)));
	connect(demEdi, SIGNAL(slideshowOn(bool)), this, SLOT(lockBenchSlot(bool)));

	// Bring on function editor finally
	functionSlot();
}


/* ********************************************************************** */
/* REIMPLEMENT EVENTS FOR LOCALIZATIONS                                   */
/* ********************************************************************** */
void MaiWindow::changeEvent(QEvent* event) {

	// Check for language event
	if(event->type() == QEvent::LanguageChange) {

		// Main stuff
		Ui_mainCtrl::retranslateUi(this);
		reTranslate();
		setWindowTitle(currentWindowName);

		// Help
		if(hlpEdi->isVisible()) {
			hlpEdi->close();
			helpSlot();
		}

		// Solver editor
		if(solEdi->isVisible())
			solEdi->setWindowTitle(tr("Solver report"));
	}
	else
		QMainWindow::changeEvent(event);
}


/* ********************************************************************** */
/* ACTIONS FOR MAINWINDOW-MENUS                                           */
/* ********************************************************************** */
void MaiWindow::newSlot() {

	oglWid->iniAll();
	for(int i=0; i<MAXFUN; i++) {
		q2cstrcpy(fun[i].str, "\0"); ofun[i]="\0";
		q2cstrcpy(iso[i].str, "\0"); oiso[i]="\0";
		q2cstrcpy(par.str[i], "\0"); opar[i]="\0";
	}

	Draw_Which();
	initAllParsers();
	updMaiWid();
	updMaiInf();

	funEdi->updFunWid();
	ligEdi->updLigWid();
	matEdi->updMatWid();
	entEdi->updEntWid();
	aniEdi->animate(false);
	aniEdi->updAniWid();
	morEdi->morph(false);
	morEdi->updMorWid();
	speEdi->updSpeWid();
	usrEdi->updUsrWid();

	setCurrentFile("");		// Window name = untitled
	emit gotNewFile();		// Draw z=0 finally
}

void MaiWindow::openSlot() {

	// File selector
	QString fileName=QFileDialog::getOpenFileName(
					this,
					tr("Open")+" "+NAME+"-"+tr("file"),
					workDir,
					"Images (*.zhu)");

	// Got file or canceled file-selector?
	if(fileName.isEmpty())
		return;

	// Check ending first
	if(!fileName.endsWith(".zhu"))
		fileName+=".zhu";

	// Load file then
	if(loadFile(fileName))
		performOGLUpdate();
	else {
		QMessageBox::critical(
			this,
			tr("Open"),
			tr("Error loading file:")+" "+fileName
		);
		return;
	}

	// Show function editor & viewer anyhow
	funEdi->showNormal();
	funEdi->raise();
	funEdi->activateWindow();
	oglWid->showNormal();
	oglWid->raise();
	oglWid->activateWindow();

	// Check animation-state at last
	oglWid->updWinTit();
}

void MaiWindow::saveAsSlot() {

	// File selector
	QString fileName=QFileDialog::getSaveFileName(
					this,
					tr("Save")+" "+NAME+"-"+tr("file"),
					workDir,
					"Images (*.zhu)");

	// Got file or canceled file-selector?
	if(fileName.isEmpty())
		return;

	// Check ending
	if(!fileName.endsWith(".zhu"))
		fileName+=".zhu";

	// Save finally
	if(!saveFile(fileName)) {
		QMessageBox::critical(this,
						tr("Save"),
						tr("Error saving file:")+" "+fileName
		);
		return;
	}
}

// Save content of solver editor
void MaiWindow::saveSolSlot() {
	const QString tmp=solEdi->toPlainText();

	// Something to save?
	if(tmp=="") {
		QMessageBox::information(
					this,
					tr("Solver report"),
					tr("Currently there is nothing to save!")
		);
		return;
	}

	// File selector
	QString fileName=QFileDialog::getSaveFileName(
					this,
					tr("Save solutions"),
					workDir,
					"Images (*.txt)");

	// Got file or canceled file-selector?
	if(fileName.isEmpty())
		return;

	// Check ending
	if(!fileName.endsWith(".txt"))
		fileName+=".txt";

	// Check success
	QFile file(fileName);
	if(!file.open(QFile::WriteOnly | QFile::Text)) {
		QMessageBox::critical(
					this,
					tr("Save"),
					tr("Error saving file:")+" "+fileName
		);
		return;
	}

	// Write and close
	QTextStream out(&file);
	out << tmp;
	file.close();
}

// Consider localized help-files too
void MaiWindow::helpSlot() {

	// Extract localized html-filename
	QString tmp=NAME;
	tmp=tmp.toLower()+"_"+zhuLocale;
	tmp=MYDOCDIR+tmp+".html";
	QFile file(tmp);

	// Try to find localized help-version
	if(!file.exists()) {

		// Not found. Try the english zhu3d_en.html instead
		tmp=NAME;
		tmp=MYDOCDIR+tmp.toLower()+"_en.html";
		file.setFileName(tmp);

		// Oops. No help-file at all
		if(!file.exists()) {
			QMessageBox::warning(
						this,
						NAME,
						tr("Cannot find any helpfile"));
			return;
		}
	}

	#ifdef WIN32
	QUrl url(QString("file:///")+tmp);
	#else
	QUrl url(QString("file://")+tmp);
	#endif

	// Use external browser ?
	if(extBrowser) {
		if(hlpEdi->isVisible())
			hlpEdi->close();
		QDesktopServices::openUrl(url);
		return;
	}

	// Else use internal browser
	hlpEdi->setWindowTitle((QString)NAME+" "+tr("Help"));
	hlpEdi->setSource(url);
	hlpEdi->show();
	return;
}

// Close all windows on exit
void MaiWindow::exitSlot() {
	writeSettings();
	oglWid->close();
	ligEdi->close();
	matEdi->close();
	entEdi->close();
	aniEdi->close();
	morEdi->close(); delete morEdi;
	picEdi->close();
	fntEdi->close();
	demEdi->close();
	hlpEdi->close();
	solEdi->close();
	funEdi->close();
	speEdi->close();
	dirEdi->close();
	usrEdi->close();
	close();
}


/* ********************************************************************** */
/* SYSTEM INFO                                                            */
/* ********************************************************************** */
void MaiWindow::systemSlot() {
	QMessageBox::information(this, tr("System information"), systemInfo());
}

void MaiWindow::aboutSlot() {

	// Get compilation date
	const QString tmp=(QString)(__DATE__);

	QMessageBox::about(this,
		tr("About")+" "+NAME,

		// Never translated
		QString("<b>")+NAME+QString("</b> ")+
        QString("is a an OpenGL-based function-viewer and solver.")+
		QString("<p>Version ")+VERS+QString("  compiled ")+tmp+QString("</p>")+
        QString("<p>Check updates on: "
		"<a href=\"http://www.kde-apps.org/content/show.php?content=43071/\""
		">www.kde-apps.org</a><BR>"

        "Contact address: "
		"<a href=\"mailto:zhu3d@aon.at?subject=Zhu3D\""
		">zhu3d@aon.at</a></p>")+

		QString("<p>(c) 1991-2008 Heinz van Saanen</p>"));
}


/* ********************************************************************** */
/* SHOW EDITORS                                                           */
/* ********************************************************************** */
void MaiWindow::lightSlot() {

	// Not yet activated?
	if(!ligEdi->isVisible())
		ligEdi->show();
	else
		ligEdi->showNormal();

	// Update and bring to foreground
	ligEdi->updLigWid();
	ligEdi->raise();
	ligEdi->activateWindow();
}

void MaiWindow::materialSlot() {

	// Not yet activated?
	if(!matEdi->isVisible())
		matEdi->show();
	else
		matEdi->showNormal();

	// Update and bring to foreground
	matEdi->updMatWid();
	matEdi->raise();
	matEdi->activateWindow();
}

void MaiWindow::animateSlot() {

	// Not yet activated?
	if(!aniEdi->isVisible())
		aniEdi->show();
	else
		aniEdi->showNormal();

	// Update and bring to foreground
	aniEdi->updAniWid();
	aniEdi->raise();
	aniEdi->activateWindow();
}

void MaiWindow::morphSlot() {

	// Not yet activated?
	if(!morEdi->isVisible())
		morEdi->show();
	else
		morEdi->showNormal();

	// Update and bring to foreground
	morEdi->updMorWid();
	morEdi->raise();
	morEdi->activateWindow();
}

void MaiWindow::entireSlot() {

	// Not yet activated?
	if(!entEdi->isVisible())
		entEdi->show();
	else
		entEdi->showNormal();

	// Update and bring to foreground
	entEdi->updEntWid();
	entEdi->raise();
	entEdi->activateWindow();
}

void MaiWindow::solverSlot() {

	// Not yet activated?
	if(!solEdi->isVisible())
		solEdi->show();
	else
		solEdi->showNormal();

	// Update and bring to foreground
	solEdi->raise();
	solEdi->activateWindow();
}

void MaiWindow::functionSlot() {

	// Not yet activated?
	if(!funEdi->isVisible())
		funEdi->show();
	else
		funEdi->showNormal();

	// Update and bring to foreground
	funEdi->updFunWid();
	funEdi->raise();
	funEdi->activateWindow();
}

void MaiWindow::pictureSlot() {

	// Not yet activated?
	if(!picEdi->isVisible()) {
		picEdi->show();
		picEdi->updPicWid();
	}
	else
		picEdi->showNormal();

	// Update and bring to foreground
	picEdi->raise();
	picEdi->activateWindow();
}

void MaiWindow::fontSlot() {

	// Not yet activated?
	if(!fntEdi->isVisible())
		fntEdi->show();
	else
		fntEdi->showNormal();

	// Update and bring to foreground
	fntEdi->raise();
	fntEdi->activateWindow();
}

void MaiWindow::demoSlot() {

	// Not yet activated?
	if(!demEdi->isVisible()) {
		demEdi->initDirList();
		demEdi->show();
	}
	else {
		demEdi->initDirList();
		demEdi->showNormal();
	}

	// Update and bring to foreground
	demEdi->raise();
	demEdi->activateWindow();
}

void MaiWindow::specialSlot() {

	// Not yet activated?
	if(!speEdi->isVisible())
		speEdi->show();
	else
		speEdi->showNormal();

	// Update and bring to foreground
	speEdi->raise();
	speEdi->activateWindow();
}

void MaiWindow::directorySlot() {

	// Not yet activated?
	if(!dirEdi->isVisible())
		dirEdi->show();
	else
		dirEdi->showNormal();

	// Update and bring to foreground
	dirEdi->raise();
	dirEdi->activateWindow();
}

void MaiWindow::userSlot() {

	// Not yet activated?
	if(!usrEdi->isVisible())
		usrEdi->show();
	else
		usrEdi->showNormal();

	// Update and bring to foreground
	usrEdi->raise();
	usrEdi->activateWindow();
}

// Bring all opened/minimized editors to front
void MaiWindow::toFrontSlot() {

	// Viewer first for not overlapping the rest
	oglWid->showNormal();
	oglWid->raise();
	oglWid->activateWindow();

	// Lights
	if(ligEdi->isVisible()) {
		ligEdi->showNormal();
		ligEdi->raise();
		ligEdi->activateWindow();
	}

	// Materials
	if(matEdi->isVisible()) {
		matEdi->showNormal();
		matEdi->raise();
		matEdi->activateWindow();
	}

	// Global light
	if(entEdi->isVisible()) {
		entEdi->showNormal();
		entEdi->raise();
		entEdi->activateWindow();
	}

	// Animation
	if(aniEdi->isVisible()) {
		aniEdi->showNormal();
		aniEdi->raise();
		aniEdi->activateWindow();
	}

	// Morphing
	if(morEdi->isVisible()) {
		morEdi->showNormal();
		morEdi->raise();
		morEdi->activateWindow();
	}

	// Special
	if(speEdi->isVisible()) {
		speEdi->showNormal();
		speEdi->raise();
		speEdi->activateWindow();
	}

	// Solver report
	if(solEdi->isVisible()) {
		solEdi->showNormal();
		solEdi->raise();
		solEdi->activateWindow();
	}

	// Function editor
	if(funEdi->isVisible()) {
		funEdi->showNormal();
		funEdi->raise();
		funEdi->activateWindow();
	}

	// User table
	if(usrEdi->isVisible()) {
		usrEdi->showNormal();
		usrEdi->raise();
		usrEdi->activateWindow();
	}
}


// Hide/minimize all opened editors
void MaiWindow::toBackSlot() {

	// Lights
	if(ligEdi->isVisible())
		ligEdi->showMinimized();

	// Materials
	if(matEdi->isVisible())
		matEdi->showMinimized();

	// Global light
	if(entEdi->isVisible())
		entEdi->showMinimized();

	// Animation
	if(aniEdi->isVisible())
		aniEdi->showMinimized();

	// Morphing
	if(morEdi->isVisible())
		morEdi->showMinimized();

	// Special
	if(speEdi->isVisible())
		speEdi->showMinimized();

	// Solver report
	if(solEdi->isVisible())
		solEdi->showMinimized();

	// Function editor
	if(funEdi->isVisible())
		funEdi->showMinimized();

	// User table
	if(usrEdi->isVisible())
		usrEdi->showMinimized();
}


// Open all editors
void MaiWindow::showAllSlot() {

	// Lights
	ligEdi->showNormal();
	ligEdi->raise();
	ligEdi->activateWindow();

	// Materials
	matEdi->showNormal();
	matEdi->raise();
	matEdi->activateWindow();

	// Global light
	entEdi->showNormal();
	entEdi->raise();
	entEdi->activateWindow();

	// Animation
	aniEdi->showNormal();
	aniEdi->raise();
	aniEdi->activateWindow();

	// Morphing
	morEdi->showNormal();
	morEdi->raise();
	morEdi->activateWindow();

	// Special
	speEdi->showNormal();
	speEdi->raise();
	speEdi->activateWindow();

	// Solver report
	solEdi->showNormal();
	solEdi->raise();
	solEdi->activateWindow();

	// Function editor
	funEdi->showNormal();
	funEdi->raise();
	funEdi->activateWindow();

	// User table
	usrEdi->showNormal();
	usrEdi->raise();
	usrEdi->activateWindow();
}


/* ********************************************************************** */
/* INITIALIZATIONS                                                        */
/* ********************************************************************** */
void MaiWindow::resetViewSlot() {

	sta.angle=35.0f;
	sta.rotX=sta.rotY=sta.rotZ=0.0f;
	sta.transX=sta.transY=0.0f;
	sta.transZ=-5.0f;
	sta.crossX=sta.crossY=0.5f;

	updMaiWid();		// Update MaiWindow widget
	updMaiInf();		// Update cross-section
	emit gotNewFile();	// Update viewer
}

// Initialize default scaling values
void MaiWindow::resetScaleSlot() {

	sta.sa=sta.sx=sta.sy=sta.sz=1.0f;
	sta.spx=sta.spy=sta.spz=1.0f;

	updMaiWid();		// Update MaiWindow widget
	updMaiInf();		// Update cross-section
	emit gotNewFile();	// Update viewer
}

// Initialize default light values
void MaiWindow::resetLightSlot() {

	iniLig();

	updMaiWid();			// Update MaiWindow widget
	updMaiInf();			// Update cross-section
	ligEdi->updLigWid();	// Update light editor
	emit gotNewFile();		// Update viewer
}

// Initialize default scaling and viewing values
void MaiWindow::resetAllSlot() {

	resetViewSlot();
	resetScaleSlot();
	resetLightSlot();
}

// Toggle animation on/off
void MaiWindow::runStopSlot() {

	ani.active = !ani.active;

	if(ani.active) {
		runStopAction->setIcon(QIcon(":/images/stop.png"));
		runStopAction->setToolTip(tr("Stop animation"));
		aniEdi->animate(true);
	}
	else {
		runStopAction->setIcon(QIcon(":/images/play.png"));
		runStopAction->setToolTip(tr("Start animation"));
		aniEdi->animate(false);
	}

	// Update OpenGL window title
	oglWid->updWinTit();
}


/* ********************************************************************** */
/* CREATE ACTIONS                                                         */
/* ********************************************************************** */
void MaiWindow::createActions() {

	// FILE
	newAction = new QAction(QIcon(":/images/new.png"), tr("&New"), this);
	newAction->setShortcut(tr("Ctrl+N"));
	connect(newAction, SIGNAL(triggered()), this, SLOT(newSlot()));

	openAction = new QAction(QIcon(":/images/open.png"), tr("&Open ..."), this);
	openAction->setShortcut(tr("Ctrl+O"));
	connect(openAction, SIGNAL(triggered()), this, SLOT(openSlot()));

	saveAsAction = new QAction(QIcon(":/images/saveas.png"), tr("&Save as"), this);
	saveAsAction->setShortcut(tr("Ctrl+S"));
	connect(saveAsAction, SIGNAL(triggered()), this, SLOT(saveAsSlot()));

	saveSolAction = new QAction(QIcon(":/images/solve.png"), tr("Save solutions"), this);
	connect(saveSolAction, SIGNAL(triggered()), this, SLOT(saveSolSlot()));

	savePicAction = new QAction(QIcon(":/images/savepic.png"), tr("Save picture"), this);
	connect(savePicAction, SIGNAL(triggered()), oglWid, SLOT(savePic()));

	printAction = new QAction(QIcon(":/images/print.png"), tr("Print picture"), this);
	connect(printAction, SIGNAL(triggered()), oglWid, SLOT(printPic()));

	exitAction = new QAction(QIcon(":/images/exit.png"), tr("&Exit"), this);
	exitAction->setShortcut(tr("Ctrl+E"));
	connect(exitAction, SIGNAL(triggered()), this, SLOT(exitSlot()));


	// EDITORS
	lightAction = new QAction(QIcon(":/images/light.png"), tr("Lights"), this);
	connect(lightAction, SIGNAL(triggered()), this, SLOT(lightSlot()));

	materialAction = new QAction(QIcon(":/images/material.png"), tr("Surfaces"), this);
	connect(materialAction, SIGNAL(triggered()), this, SLOT(materialSlot()));

	entireAction = new QAction(QIcon(":/images/entire.png"), tr("Global"), this);
	connect(entireAction, SIGNAL(triggered()), this, SLOT(entireSlot()));

	animateAction = new QAction(QIcon(":/images/play.png"), tr("Animation"), this);
	connect(animateAction, SIGNAL(triggered()), this, SLOT(animateSlot()));

	morphAction = new QAction(QIcon(":/images/morph.png"), tr("Morphing"), this);
	connect(morphAction, SIGNAL(triggered()), this, SLOT(morphSlot()));

	solEdiAction = new QAction(QIcon(":/images/solve.png"), tr("Solver report"), this);
	connect(solEdiAction, SIGNAL(triggered()), this, SLOT(solverSlot()));

	usrEdiAction = new QAction(QIcon(":/images/user.png"), tr("User defined"), this);
	connect(usrEdiAction, SIGNAL(triggered()), this, SLOT(userSlot()));

	funEdiAction = new QAction(QIcon(":/images/edit.png"), tr("Functions"), this);
	connect(funEdiAction, SIGNAL(triggered()), this, SLOT(functionSlot()));


	// SETTINGS
	pictureAction = new QAction(QIcon(":/images/savepic.png"), tr("Picture"), this);
	connect(pictureAction, SIGNAL(triggered()), this, SLOT(pictureSlot()));

	sfineAction = new QAction(tr("micro"), this);
	sfineAction->setCheckable(true);
	sfineAction->setChecked(false);
	connect(sfineAction, SIGNAL(triggered()), oglWid, SLOT(sfineScale()));
	connect(sfineAction, SIGNAL(triggered()), this,  SLOT(sfineScale()));

	fineAction = new QAction(tr("fine"), this);
	fineAction->setCheckable(true);
	fineAction->setChecked(false);
	connect(fineAction, SIGNAL(triggered()), oglWid, SLOT(fineScale()));
	connect(fineAction, SIGNAL(triggered()), this,  SLOT(fineScale()));

	normalAction = new QAction(tr("normal"), this);
	normalAction->setCheckable(true);
	normalAction->setChecked(true);
	connect(normalAction, SIGNAL(triggered()), oglWid, SLOT(normalScale()));
	connect(normalAction, SIGNAL(triggered()), this,  SLOT(normalScale()));

	coarseAction = new QAction(tr("coarse"), this);
	coarseAction->setCheckable(true);
	coarseAction->setChecked(false);
	connect(coarseAction, SIGNAL(triggered()), oglWid, SLOT(coarseScale()));
	connect(coarseAction, SIGNAL(triggered()), this,  SLOT(coarseScale()));

	resetViewAction = new QAction(QIcon(":/images/viewonly.png"), tr("View only"), this);
	connect(resetViewAction, SIGNAL(triggered()), this, SLOT(resetViewSlot()));

	resetScaleAction = new QAction(QIcon(":/images/scale.png"), tr("Scaling only"), this);
	connect(resetScaleAction, SIGNAL(triggered()), this, SLOT(resetScaleSlot()));

	resetLightAction = new QAction(QIcon(":/images/light.png"), tr("Lights only"), this);
	connect(resetLightAction, SIGNAL(triggered()), this, SLOT(resetLightSlot()));

	resetAllAction = new QAction(QIcon(":/images/reset.png"), tr("All together"), this);
	connect(resetAllAction, SIGNAL(triggered()), this, SLOT(resetAllSlot()));

	fontAction = new QAction(QIcon(":/images/icon.png"), tr("General"), this);
	connect(fontAction, SIGNAL(triggered()), this, SLOT(fontSlot()));

	speEdiAction = new QAction(QIcon(":/images/special.png"), tr("Special"), this);
	connect(speEdiAction, SIGNAL(triggered()), this, SLOT(specialSlot()));

	dirEdiAction = new QAction(QIcon(":/images/open.png"), tr("Directories"), this);
	connect(dirEdiAction, SIGNAL(triggered()), this, SLOT(directorySlot()));


	// WINDOWS
	toFrontAction = new QAction(QIcon(":/images/tofront.png"), tr("Show open editors"), this);
	connect(toFrontAction, SIGNAL(triggered()), this, SLOT(toFrontSlot()));

	toBackAction = new QAction(QIcon(":/images/toback.png"), tr("Hide open editors"), this);
	connect(toBackAction, SIGNAL(triggered()), this, SLOT(toBackSlot()));

	showAllAction = new QAction(QIcon(":/images/edit.png"), tr("Open all editors"), this);
	connect(showAllAction, SIGNAL(triggered()), this, SLOT(showAllSlot()));


	// HELP
	helpAction = new QAction(QIcon(":/images/help.png"), tr("Help"), this);
	helpAction->setShortcut(tr("F1"));
	connect(helpAction, SIGNAL(triggered()), this, SLOT(helpSlot()));

	systemAction = new QAction(QIcon(":/images/system.png"), tr("System"), this);
	connect(systemAction, SIGNAL(triggered()), this, SLOT(systemSlot()));

	demoAction = new QAction(QIcon(":/images/demo.png"), tr("Demo"), this);
	connect(demoAction, SIGNAL(triggered()), this, SLOT(demoSlot()));

	benchmarkAction = new QAction(QIcon(":/images/bench.png"), tr("Benchmark"), this);
	connect(benchmarkAction, SIGNAL(triggered()), this, SLOT(benchmarkSlot()));

	aboutAction = new QAction(QIcon(":/images/icon.png"), tr("About Zhu3D"), this);
	connect(aboutAction, SIGNAL(triggered()), this, SLOT(aboutSlot()));

	aboutQtAction = new QAction(QIcon(":/images/qt-logo.png"), tr("About Qt"), this);
	connect(aboutQtAction, SIGNAL(triggered()), qApp, SLOT(aboutQt()));


	// TOOLBARS
	zoomInAction = new QAction(QIcon(":/images/zoomin.png"), tr("Zoom in"), this);
	zoomInAction->setShortcut(tr("Ctrl++"));
	connect(zoomInAction, SIGNAL(triggered()), oglWid, SLOT(zoomIn()));

	zoomOutAction = new QAction(QIcon(":/images/zoomout.png"), tr("Zoom out"), this);
	zoomOutAction->setShortcut(tr("Ctrl+-"));
	connect(zoomOutAction, SIGNAL(triggered()), oglWid, SLOT(zoomOut()));

	transLeftAction = new QAction(QIcon(":/images/left.png"), tr("Move left"), this);
	transLeftAction->setShortcut(tr("Ctrl+L"));
	connect(transLeftAction, SIGNAL(triggered()), oglWid, SLOT(transLeft()));

	transRightAction = new QAction(QIcon(":/images/right.png"), tr("Move right"), this);
	transRightAction->setShortcut(tr("Ctrl+R"));
	connect(transRightAction, SIGNAL(triggered()), oglWid, SLOT(transRight()));

	transUpAction = new QAction(QIcon(":/images/up.png"), tr("Move up"), this);
	transUpAction->setShortcut(tr("Ctrl+U"));
	connect(transUpAction, SIGNAL(triggered()), oglWid, SLOT(transUp()));

	transDownAction = new QAction(QIcon(":/images/down.png"), tr("Move down"), this);
	transDownAction->setShortcut(tr("Ctrl+D"));
	connect(transDownAction, SIGNAL(triggered()), oglWid, SLOT(transDown()));

	solveAction = new QAction(QIcon(":/images/solve.png"), tr("Solve system"), this);
	connect(solveAction, SIGNAL(triggered()), this, SLOT(solveSlot()));

	runStopAction = new QAction(QIcon(":/images/play.png"), tr("Start/stop animation"), this);
	connect(runStopAction, SIGNAL(triggered()), this, SLOT(runStopSlot()));
}


/* ********************************************************************** */
/* CREATE/DELETE MENUS                                                    */
/* ********************************************************************** */
void MaiWindow::createMenus() {

	// File
	fileMenu = menuBar()->addMenu(tr("File"));
	fileMenu->addAction(newAction);
	fileMenu->addAction(openAction);
	fileMenu->addAction(saveAsAction);
	fileMenu->addSeparator();
	fileMenu->addAction(saveSolAction);
	fileMenu->addAction(savePicAction);
	fileMenu->addAction(printAction);
	fileMenu->addSeparator();
	fileMenu->addAction(exitAction);

	// Editors
	editorMenu = menuBar()->addMenu(tr("Editors"));
	editorMenu->addAction(lightAction);
	editorMenu->addAction(materialAction);
	editorMenu->addAction(entireAction);
	editorMenu->addSeparator();
	editorMenu->addAction(animateAction);
	editorMenu->addAction(morphAction);
	editorMenu->addSeparator();
	editorMenu->addAction(solEdiAction);
	editorMenu->addAction(usrEdiAction);
	editorMenu->addSeparator();
	editorMenu->addAction(funEdiAction);

	// Settings with submenus
	settingsMenu = menuBar()->addMenu(tr("Settings"));

	scaleMenu = settingsMenu->addMenu(QIcon(":/images/scale.png"), tr("Scaling"));
	scaleMenu->addAction(sfineAction);
	scaleMenu->addAction(fineAction);
	scaleMenu->addAction(normalAction);
	scaleMenu->addAction(coarseAction);

	resetMenu = settingsMenu->addMenu(QIcon(":/images/reset.png"), tr("Reset"));
	resetMenu->addAction(resetViewAction);
	resetMenu->addAction(resetScaleAction);
	resetMenu->addAction(resetLightAction);
	resetMenu->addSeparator();
	resetMenu->addAction(resetAllAction);

	settingsMenu->addSeparator();
	settingsMenu->addAction(fontAction);
	settingsMenu->addAction(dirEdiAction);
	settingsMenu->addAction(pictureAction);
	settingsMenu->addSeparator();
	settingsMenu->addAction(speEdiAction);

	// Windows
	windowsMenu = menuBar()->addMenu(tr("Windows"));
	windowsMenu->addAction(toFrontAction);
	windowsMenu->addAction(toBackAction);
	windowsMenu->addSeparator();
	windowsMenu->addAction(showAllAction);

	// Help
	helpMenu = menuBar()->addMenu(tr("Help"));
	helpMenu->addAction(helpAction);
	helpMenu->addAction(systemAction);
	helpMenu->addSeparator();
	helpMenu->addAction(demoAction);
	helpMenu->addAction(benchmarkAction);
	helpMenu->addSeparator();
	helpMenu->addAction(aboutAction);
	helpMenu->addAction(aboutQtAction);
}


/* ********************************************************************** */
/* CREATE TOOLBARS                                                        */
/* ********************************************************************** */
void MaiWindow::createToolBars() {

	// Get current iconsize from fntEdi
	QSize iconSize=fntEdi->getIcoSize();

	fileToolBar = addToolBar(tr("File"));
	fileToolBar->addAction(openAction);
	fileToolBar->addAction(saveAsAction);
	fileToolBar->setIconSize(iconSize);

	zoomToolBar = addToolBar(tr("Zoom"));
	zoomToolBar->addAction(zoomInAction);
	zoomToolBar->addAction(zoomOutAction);
	zoomToolBar->setIconSize(iconSize);

	transToolBar = addToolBar(tr("Move"));
	transToolBar->addAction(transLeftAction);
	transToolBar->addAction(transRightAction);
	transToolBar->addAction(transUpAction);
	transToolBar->addAction(transDownAction);
	transToolBar->setIconSize(iconSize);

	extraToolBar = addToolBar(tr("Extras"));
	extraToolBar->addAction(solveAction);
	extraToolBar->addAction(runStopAction);
	extraToolBar->setIconSize(iconSize);
}

// Like in menu-redrawing, again strange workaround is needed!!!
void MaiWindow::iconSizeSlot(const int newSize) {
	QSize iconSize(QSize(newSize, newSize));

	// Destroy toolbars first
	delete fileToolBar;
	delete zoomToolBar;
	delete transToolBar;
	delete extraToolBar;

	// Recrate them
	createToolBars();

	// And now resize them!
	fileToolBar->setIconSize(iconSize);
	zoomToolBar->setIconSize(iconSize);
	transToolBar->setIconSize(iconSize);
	extraToolBar->setIconSize(iconSize);
}


/* ********************************************************************** */
/* MISCELLANEOUS GOODIES                                                  */
/* ********************************************************************** */
void MaiWindow::readSettings() {

	QSettings settings("Heinz", NAME);
	QPoint pos = settings.value("MainWindowPos", QPoint(200, 200)).toPoint();
	QSize size = settings.value("MainWindowSize", QSize(400, 400)).toSize();
	QPoint solpos = settings.value("SolverEditPos", QPoint(200, 200)).toPoint();
	QSize solsize = settings.value("SolverEditSize", QSize(400, 400)).toSize();
	QPoint hlppos = settings.value("HelpEditPos", QPoint(200, 200)).toPoint();
	QSize hlpsize = settings.value("HelpEditSize", QSize(400, 400)).toSize();

	// Ensure to start with defaults on 1st program-start ever
	everStarted=false;
	everStarted=settings.value("FirstMainStart").toBool();
	if(everStarted) {
		resize(size);
		move(pos);
		solEdi->resize(solsize);
		solEdi->move(solpos);
		hlpEdi->resize(hlpsize);
		hlpEdi->move(hlppos);
	}
	else {
		everStarted=true;

		// Check clockcounter
		if(MHz==0.0f)
			QMessageBox::information(
				0,
				QObject::tr("Program start")+"\n",
				QObject::tr("CPU clockcounter is not supported!")+"\n\n"+
				QObject::tr("System info cannot provide MHz-values")
			);

		// Check hardware acceleration
		if(!oglhw_avail())
			QMessageBox::information(
				0,
				QObject::tr("Program start")+"\n",
				QObject::tr("Missing OpenGL acceleration!")+"\n\n"+
				QObject::tr("Framerates will be very low")
		);

		resize(QSize(350,440));
		move(QPoint(10, 10));
		solEdi->resize(QSize(700, 320));
		solEdi->move(QPoint(300, 350));
		hlpEdi->resize(QSize(600, 680));
		hlpEdi->move(QPoint(40, 50));
	}
}

void MaiWindow::writeSettings() {
	QSettings settings("Heinz", NAME);

	// Save actual settings
	settings.setValue("MainWindowPos", pos());
	settings.setValue("MainWindowSize", size());
	settings.setValue("SolverEditPos", solEdi->pos());
	settings.setValue("SolverEditSize", solEdi->size());
	settings.setValue("HelpEditPos", hlpEdi->pos());
	settings.setValue("HelpEditSize", hlpEdi->size());
	settings.setValue("FirstMainStart", everStarted);
}

// Wire-mode+Points is not sensefull, so test if wired is checkable
void MaiWindow::wireModeSlot() {

	// mode==point ?
	if(sta.draw==2) {
		Wired->setChecked(false);
		Wired->setEnabled(false);
		return;
	}

	Wired->setChecked(sta.wired);
	Wired->setEnabled(true);
}

// Solver in oglWid has done his work
void MaiWindow::solveSlot() {
	const QDate date=QDate::currentDate();
	const QTime time=QTime::currentTime();
	const QString tmp=solve();

	// Solver successful?
	if(tmp.isEmpty())
		return;

	// Fill solver editor
	solEdi->append("********************************************************************************");
	solEdi->append(tr("Creation date:")+"  "+date.toString()+"  "+time.toString());
	solEdi->append("********************************************************************************");
	solEdi->append("F0(X,Y) = " + ofun[0]);
	solEdi->append("F1(X,Y) = " + ofun[1]);
	solEdi->append("F2(X,Y) = " + ofun[2]);
	solEdi->append("");
	solEdi->append(tmp);
	solEdi->append("********************************************************************************");
	solEdi->append("\n");

	// Update cross-positions
	updMaiInf();
	emit gotNewView();

	// Set cursor at end
	QTextCursor textCursor;
	textCursor=solEdi->textCursor();
	textCursor.movePosition(QTextCursor::End);
	solEdi->setTextCursor(textCursor);

	// Solver editor to foreground
	solverSlot();
}

// Show cross statistics
void MaiWindow::updMaiInf() {
	QString message;
	QString tmp;
	const int crossPrec=6;
	float val, f;

	// Prepare
	infoTextEdit->clear();
	infoTextEdit->setFontItalic(false);
	infoTextEdit->setTextColor(QColor(0,0,0));

	// Current xy-value for actual cross-position
	if(sta.funMod!=PARAMETER) {

		val=fabsf(f=sta.crossX*sta.sx);
		if(val>100000.0f || val<0.00001f)
			message=tr("Cross")+" X = "+tmp.setNum(f, 'e', crossPrec);
		else
			message=tr("Cross")+" X = "+tmp.setNum(f, 'f', crossPrec);
		infoTextEdit->append(message);

		val=fabsf(f=sta.crossY*sta.sy);
		if(val>100000.0f || val<0.00001f)
			message=tr("Cross")+" Y = "+tmp.setNum(f, 'e', crossPrec);
		else
			message=tr("Cross")+" Y = "+tmp.setNum(f, 'f', crossPrec);
		message+="\n------------------------------";
		}
	else {

		val=fabsf(f=sta.crossX*sta.spx/3.141596f/sta.spz);
		if(val>100000.0f || val<0.00001f)
			message=tr("Cross")+" X = "+tmp.setNum(f, 'e', crossPrec);
		else
			message=tr("Cross")+" X = "+tmp.setNum(f, 'f', crossPrec);
		infoTextEdit->append(message);

		val=fabsf(f=sta.crossY*sta.spy/3.141596f/sta.spz);
		if(val>100000.0f || val<0.00001f)
			message=tr("Cross")+" Y = "+tmp.setNum(f, 'e', crossPrec);
		else
			message=tr("Cross")+" Y = "+tmp.setNum(f, 'f', crossPrec);
		message+="\n------------------------------";
	}
	infoTextEdit->append(message);

	// Change according to mode
	if(sta.funMod!=FUNCTION) {
		infoTextEdit->setFontItalic(true);
		infoTextEdit->setTextColor(QColor(210,210,210));
	}

	// Calculate and show function values at cross-position
	sta.crossZ=fxy0(sta.crossX, sta.crossY)/sta.sz;
	if(fabs(sta.crossZ*sta.sx)>100000.0f || fabs(sta.crossZ*sta.sx)<0.00001f)
		message="F0(X,Y) = "+tmp.setNum(sta.crossZ, 'e', crossPrec);
	else
		message="F0(X,Y) = "+tmp.setNum(sta.crossZ, 'f', crossPrec);
	infoTextEdit->append(message);

	sta.crossZ=fxy1(sta.crossX, sta.crossY)/sta.sz;
	if(fabs(sta.crossZ*sta.sx)>100000.0f || fabs(sta.crossZ*sta.sx)<0.00001f)
		message="F1(X,Y) = "+tmp.setNum(sta.crossZ, 'e', crossPrec);
	else
		message="F1(X,Y) = "+tmp.setNum(sta.crossZ, 'f', crossPrec);
	infoTextEdit->append(message);

	sta.crossZ=fxy2(sta.crossX, sta.crossY)/sta.sz;
	if(fabs(sta.crossZ*sta.sx)>100000.0f || fabs(sta.crossZ*sta.sx)<0.00001f)
		message="F2(X,Y) = "+tmp.setNum(sta.crossZ, 'e', crossPrec);
	else
		message="F2(X,Y) = "+tmp.setNum(sta.crossZ, 'f', crossPrec);
	infoTextEdit->append(message);
}


/* ********************************************************************** */
/* FILE STUFF                                                             */
/* ********************************************************************** */

// Load
bool MaiWindow::loadFile(const QString &fileName) {
	long i, j, k;
	QFile file(fileName);
	QString tmp;

	// Open with checks
	if(!file.exists() || !file.open(QFile::ReadOnly | QFile::Text)) {
		QMessageBox::critical(this,
					tr("Open"),
					tr("Error loading file:")+" "+fileName
		);
		return false;
	}

	// Cursor is resetted again after ogl-drawing
	QTextStream in(&file);
	QApplication::setOverrideCursor(Qt::WaitCursor);

	// Check header
	tmp=file.readLine(); tmp.remove(QChar('\n'));
	if(!tmp.startsWith("Zhu3D")) {
		QApplication::restoreOverrideCursor();
		file.close();
			QMessageBox::warning(this,
						tr("Open"),
						tr("Zhu3D file is too old or not valid!")
			);
			return false;
	}

	// Stop a running animation
	if(ani.active)
		aniEdi->animate(false);

	// Clean up whole screen before loading new data
	for(i=0; i<MAXFUN; i++)
		fun[i].drawFun=false;
	par.drawPar=sta.axes=sta.cross=false;
	sta.funMod=FUNCTION;
	Draw_Which();
	emit gotNewView();

	// Table/error/morph inits
	iniErr();
	iniUsr();
	iniMor();

	// Get functions & parametric system
	for(i=0; i<MAXFUN; i++) { ofun[i]=file.readLine(); ofun[i].remove(QChar('\n')); if(ofun[i]=="@") ofun[i]="\0"; }
	for(i=0; i<MAXFUN; i++) { oiso[i]=file.readLine(); oiso[i].remove(QChar('\n')); if(oiso[i]=="@") oiso[i]="\0"; }
	for(i=0; i<MAXFUN; i++) { opar[i]=file.readLine(); opar[i].remove(QChar('\n')); if(opar[i]=="@") opar[i]="\0"; }

	// Prepare for proper parsing
	for(i=0; i<MAXFUN; i++) { tmp=ofun[i]; q2cstrcpy(fun[i].str, prepareInput(tmp)); }
	for(i=0; i<MAXFUN; i++) { tmp=oiso[i]; q2cstrcpy(iso[i].str, prepareInput(tmp)); }
	for(i=0; i<MAXFUN; i++) { tmp=opar[i]; q2cstrcpy(par.str[i], prepareInput(tmp)); }

	// Get status
	for(i=0; i<MAXFUN; i++) { in >> fun[i].drawFun; }
	for(i=0; i<MAXFUN; i++) { in >> iso[i].drawIso; }
	in >> par.drawPar;

	in >> par.sLow >> par.sHig >> par.tLow >> par.tHig;
	in >> sta.spx >> sta.spy >> sta.spz;
	in >> sta.funMod;
	in >> sta.angle;
	in >> sta.rotX >> sta.rotY >> sta.rotZ;
	in >> sta.transX >> sta.transY >> sta.transZ;
	in >> sta.crossX >> sta.crossY;
	in >> sta.sa >> sta.sx >> sta.sy >> sta.sz;
	in >> sta.fgrids;
	in >> sta.tessWidth;
	in >> sta.tgrids;
	in >> sta.triMod;

	sta.tessWidth=2.0f/sta.fgrids;
	in >> sta.cross >> sta.axes >> sta.wired;
	in >> sta.draw;
	in >> sta.ediLig >> sta.ediMat;
	in >> sta.axisCol;
	in >> sta.comment;		// Currently not used!

	// Lights
	in >> sta.ediLig;
	for(i=0; i<MAXLIG; i++) {
		in >> lig[i].setLig;
		in >> lig[i].ambLigLoc;
		in >> lig[i].difLigLoc;
		in >> lig[i].speLigLoc;
		in >> lig[i].ligDis;
		for(j=0; j<3; j++) in >> lig[i].ligPos[j];
		for(j=0; j<4; j++) in >> lig[i].ambLig[j];
		for(j=0; j<4; j++) in >> lig[i].difLig[j];
		for(j=0; j<4; j++) in >> lig[i].speLig[j];
		for(j=0; j<4; j++) in >> lig[i].spoDir[j];
		in >> lig[i].spoSta;
		in >> lig[i].spoExp;
		in >> lig[i].spoCut;
		in >> lig[i].conAtt;
		in >> lig[i].linAtt;
		in >> lig[i].quaAtt;
	}

	// Materials
	in >> sta.ediMat;
	for(i=0; i<2; i++)
		for(j=0; j<MAXFUN+1; j++) {
			in >> mat[i].ambMatLoc[i];
			in >> mat[i].difMatLoc[i];
			in >> mat[i].speMatLoc[i];
			in >> mat[i].emiMatLoc[i];
			for(k=0; k<4; k++) in >> mat[j].ambMat[i][k];
			for(k=0; k<4; k++) in >> mat[j].difMat[i][k];
			for(k=0; k<4; k++) in >> mat[j].speMat[i][k];
			for(k=0; k<4; k++) in >> mat[j].emiMat[i][k];
			in >> mat[j].shiMat[i];
	}
	for(i=0; i<MAXFUN+1; i++)
		in >> mat[i].twoSid;

	// Entire
	in >> ent.ambEntLoc;
	in >> ent.bacEntLoc;
	in >> ent.modInf;
	in >> ent.modOne;
	for(i=0; i<4; i++) in >> ent.bacLig[i];
	for(i=0; i<4; i++) in >> ent.modLig[i];

	// Animation
	in >> ani.xFac >> ani.yFac >> ani.zFac;
	in >> ani.locXYZ;
	in >> ani.active;
	in >> ani.fps;

	// Fog
	in >> fog.enabled;
	in >> fog.fogLigLoc;
	for(i=0; i<4; i++) in >> fog.fogLig[i];
	in >> fog.fogType;
	in >> fog.intensity;
	in >> fog.start;
	in >> fog.end;

	// Textures
	for(i=0; i<MAXFUN+1; i++) {
		in >> tex[i].name;
		in >> tex[i].enabled;
		in >> tex[i].flip;
		in >> tex[i].hsize;
		in >> tex[i].vsize;
	}
	in >> texSpan;

	// Try to load all textures from current textures-directory. Even done
	// when tex[x].name=="None", because memory management is handled there
	for(int i=0; i<MAXFUN+1; i++)
		speEdi->loadTexture(texDir+tex[i].name, i);

	// Motion blur
	in >> mot.enabled;
	in >> mot.steps;
	in >> mot.fac;
	in >> mot.trans;

	// User table
	in.readLine();
	for(i=0; i<MAXUSR; i++) {
		usr[i]=in.readLine();
		if(usr[i].startsWith(QChar('+')))
			usr[i]="\0";
		else
			usrHasEntry=true;
	}

	// Morphing. One line to spare room for future extensions
	tmp=in.readLine(); tmp.remove(QChar('\n'));
	if(tmp!="+") {
		mor.active=tmp.section(' ',0,0).toInt();
		mor.lower=tmp.section(' ',1,1).toFloat();
		mor.upper=tmp.section(' ',2,2).toFloat();
		mor.steps=tmp.section(' ',3,3).toInt();
		mor.val=tmp.section(' ',4,4).toFloat();
	}

	// Room for future extensions
	for(i=0; i<35-MAXUSR; i++)
		in >> tmp;

	// Skip CR and read comment
	char cr;
	in >> cr;
	comment=in.readAll();

	// Everthing loaded. Check parser sanity now
	if(initAllParsers()!=FunctionParser::FP_NO_ERROR) {
		newSlot();
		return false;
	}

	// Perform updates
	updMaiWid();
	updMaiInf();
	funEdi->updFunWid();
	ligEdi->updLigWid();
	matEdi->updMatWid();
	entEdi->updEntWid();
	aniEdi->updAniWid();
	morEdi->updMorWid();
	speEdi->updSpeWid();
	usrEdi->updUsrWid();

	// Don't forget
	iniITL();						// Recalculate iso thread list before morphing!
	setCurrentFile(fileName);		// New window name
	aniEdi->animate(ani.active);	// Set animation mode
	morEdi->morph(mor.active);		// Set morphing mode

	// Clean ups
	PROC_EVENT;
	file.close();
	return true;
}

// Save
bool MaiWindow::saveFile(const QString &fileName) {
	long i, j, k;
	QFile file(fileName);

	// Check success
	if(!file.open(QFile::WriteOnly | QFile::Text))
		return false;

	QTextStream out(&file);
	QApplication::setOverrideCursor(Qt::WaitCursor);

	// Functions & parametric system
	out << (QString)"Zhu3D "+VERS << "\n";
	for(i=0; i<MAXFUN; i++) { if(ofun[i].isEmpty()) ofun[i]="@"; out << ofun[i] << "\n"; }
	for(i=0; i<MAXFUN; i++) { if(oiso[i].isEmpty()) oiso[i]="@"; out << oiso[i] << "\n"; }
	for(i=0; i<MAXFUN; i++) { if(opar[i].isEmpty()) opar[i]="@"; out << opar[i] << "\n"; }

	// Save Status
	for(i=0; i<MAXFUN; i++) { out << fun[i].drawFun << "\n"; }
	for(i=0; i<MAXFUN; i++) { out << iso[i].drawIso << "\n"; }
	out << par.drawPar << "\n";

	out << par.sLow << "\n" << par.sHig << "\n" << par.tLow << "\n" << par.tHig << "\n";
	out << sta.spx << "\n" << sta.spy << "\n" << sta.spz << "\n";
	out << sta.funMod << "\n";
	out << sta.angle << "\n";
	out << sta.rotX << "\n" << sta.rotY << "\n" << sta.rotZ << "\n";
	out << sta.transX << "\n" << sta.transY << "\n" << sta.transZ << "\n";
	out << sta.crossX << "\n" << sta.crossY << "\n";
	out << sta.sa << "\n" << sta.sx << "\n" << sta.sy << "\n" << sta.sz << "\n";
	out << sta.fgrids << "\n";
	out << sta.tessWidth << "\n";
	out << sta.tgrids << "\n";
	out << sta.triMod << "\n";

	out << sta.cross << "\n" << sta.axes << "\n" << sta.wired << "\n";
	out << sta.draw << "\n";
	out << sta.ediLig << "\n" << sta.ediMat << "\n";
	out << sta.axisCol << "\n";
	out << sta.comment << "\n";		// Currently not used!

	// Lights
	out << sta.ediLig << "\n";
	for(i=0; i<MAXLIG; i++) {
		out << lig[i].setLig << "\n";
		out << lig[i].ambLigLoc << "\n";
		out << lig[i].difLigLoc << "\n";
		out << lig[i].speLigLoc << "\n";
		out << lig[i].ligDis << "\n";
		for(j=0; j<3; j++) out << lig[i].ligPos[j] << "\n";
		for(j=0; j<4; j++) out << lig[i].ambLig[j] << "\n";
		for(j=0; j<4; j++) out << lig[i].difLig[j] << "\n";
		for(j=0; j<4; j++) out << lig[i].speLig[j] << "\n";
		for(j=0; j<4; j++) out << lig[i].spoDir[j] << "\n";
		out << lig[i].spoSta << "\n";
		out << lig[i].spoExp << "\n";
		out << lig[i].spoCut << "\n";
		out << lig[i].conAtt << "\n";
		out << lig[i].linAtt << "\n";
		out << lig[i].quaAtt << "\n";
	}

	// Materials
	out << sta.ediMat << "\n";;
	for(i=0; i<2; i++)
		for(j=0; j<MAXFUN+1; j++) {
			out << mat[i].ambMatLoc[i] << "\n";;
			out << mat[i].difMatLoc[i] << "\n";;
			out << mat[i].speMatLoc[i] << "\n";;
			out << mat[i].emiMatLoc[i] << "\n";;
			for(k=0; k<4; k++) out << mat[j].ambMat[i][k] << "\n";;
			for(k=0; k<4; k++) out << mat[j].difMat[i][k] << "\n";;
			for(k=0; k<4; k++) out << mat[j].speMat[i][k] << "\n";;
			for(k=0; k<4; k++) out << mat[j].emiMat[i][k] << "\n";;
			out << mat[j].shiMat[i] << "\n";;
	}
	for(i=0; i<MAXFUN+1; i++)
		out << mat[i].twoSid << "\n";

	// Entire
	out << ent.ambEntLoc << "\n";
	out << ent.bacEntLoc << "\n";
	out << ent.modInf << "\n";
	out << ent.modOne << "\n";
	for(i=0; i<4; i++) out << ent.bacLig[i] << "\n";
	for(i=0; i<4; i++) out << ent.modLig[i] << "\n";

	// Animation
	out << ani.xFac << "\n" << ani.yFac << "\n" << ani.zFac << "\n";
	out << ani.locXYZ << "\n";
	out << ani.active << "\n";
	out << ani.fps << "\n";

	// Fog
	out << fog.enabled << "\n";
	out << fog.fogLigLoc << "\n";
	for(i=0; i<4; i++) out << fog.fogLig[i] << "\n";
	out << fog.fogType << "\n";
	out << fog.intensity << "\n";
	out << fog.start << "\n";
	out << fog.end << "\n";

	// Textures; never translate file entries
	for(i=0; i<MAXFUN+1; i++) {
		if(tex[i].name!="")
			out << tex[i].name << "\n";
		else
			out << "None" << "\n";
		out << tex[i].enabled << "\n";
		out << tex[i].flip << "\n";
		out << tex[i].hsize << "\n";
		out << tex[i].vsize << "\n";
	}
	out << texSpan << "\n";

	// Motion blur
	out << mot.enabled << "\n";
	out << mot.steps << "\n";
	out << mot.fac << "\n";
	out << mot.trans << "\n";

	// User table
	for(i=0; i<MAXUSR; i++)
		if(!usr[i].isEmpty())
			out << q2cstrcpy(usr[i]) << "\n";
		else
			out << "+" << "\n";

	// Morphing. One line to spare room for future extensions
	QString tmp, low, upp, ste, val;
	tmp=tmp.setNum(mor.active) + " " + low.setNum(mor.lower,'f',6) + " " + upp.setNum(mor.upper,'f',6) + " ";
	tmp+=ste.setNum(mor.steps) + " " + val.setNum(mor.val,'f',6);
	out << q2cstrcpy(tmp) << "\n";

	// Room for future extensions
	for(i=0; i<35-MAXUSR; i++)
		out << "+" << "\n";

	// Write comment
	if((tmp=funEdi->getCom()) != "")
		out << tmp;

	// Clean up
	QApplication::restoreOverrideCursor();
	setCurrentFile(fileName);
	file.close();
	return true;
}

// Prepare filename as current window-name
void MaiWindow::setCurrentFile(const QString &fileName) {

	if(fileName.isEmpty())
		currentWindowName=tr("untitled");
	else
		currentWindowName=QFileInfo(fileName).fileName();

	currentWindowName=currentWindowName+" - "+NAME;
	setWindowTitle(currentWindowName);
}


/* ********************************************************************** */
/* HANDLE CHECKBOXES FOR SUBMENUS FILEFORMAT AND SCALING                  */
/* ********************************************************************** */
void MaiWindow::sfineScale() {
	if(sfineChecked) {
		sfineAction->setChecked(true);
		return;
	}
	sfineChecked=true;
	sfineAction->setChecked(true);
	fineChecked=false;
	fineAction->setChecked(false);
	normalChecked=false;
	normalAction->setChecked(false);
	coarseChecked=false;
	coarseAction->setChecked(false);
}

void MaiWindow::fineScale() {
	if(fineChecked) {
		fineAction->setChecked(true);
		return;
	}
	sfineChecked=false;
	sfineAction->setChecked(false);
	fineChecked=true;
	fineAction->setChecked(true);
	normalChecked=false;
	normalAction->setChecked(false);
	coarseChecked=false;
	coarseAction->setChecked(false);
}

void MaiWindow::normalScale() {
	if(normalChecked) {
		normalAction->setChecked(true);
		return;
	}
	sfineChecked=false;
	sfineAction->setChecked(false);
	fineChecked=false;
	fineAction->setChecked(false);
	normalChecked=true;
	normalAction->setChecked(true);
	coarseChecked=false;
	coarseAction->setChecked(false);
}

void MaiWindow::coarseScale() {
	if(coarseChecked) {
		coarseAction->setChecked(true);
		return;
	}
	sfineChecked=false;
	sfineAction->setChecked(false);
	fineChecked=false;
	fineAction->setChecked(false);
	normalChecked=false;
	normalAction->setChecked(false);
	coarseChecked=true;
	coarseAction->setChecked(true);
}


/* ********************************************************************** */
/* HANDLE MAIN-WIDGET REDRAW WHEN NEW FILE IS LOADED                      */
/* ********************************************************************** */
void MaiWindow::updMaiWid() {

	QString tmp;

	// Show
	Function3->setChecked(par.drawPar);
	Cross->setChecked(sta.cross);
	Axes->setChecked(sta.axes);

	// Mode
	Point->setChecked(sta.draw==2?1:0);
	Triangle->setChecked(sta.draw==0?1:0);
	Quad->setChecked(sta.draw==1?1:0);
	wireModeSlot();

	// Rotation-sliders
	Ui_mainCtrl::xSlider->setValue((int)sta.rotX);
	Ui_mainCtrl::ySlider->setValue((int)sta.rotY);
	Ui_mainCtrl::zSlider->setValue((int)sta.rotZ);

	// Scale box
	xscaleLabel->setToolTip(tr("X-values"));
	yscaleLabel->setToolTip(tr("Y-values"));
	zscaleLabel->setToolTip(tr("Z-values"));
	xyzscaleLabel->setText(tr(" A"));
	xyzscaleLabel->setToolTip(tr("All XYZ-values together"));

	// Are we in function-mode?
	if(sta.funMod==FUNCTION) {

		Function0->setText("F0");
		Function1->setText("F1");
		Function2->setText("F2");
		Function0->setChecked(fun[0].drawFun);
		Function1->setChecked(fun[1].drawFun);
		Function2->setChecked(fun[2].drawFun);
		Function0->setEnabled(true);
		Function1->setEnabled(true);
		Function2->setEnabled(true);

		ScaleBox->setToolTip(tr("Scale functions"));
		zPlus->setEnabled(true);
		zMinus->setEnabled(true);
		zscaleLabel->setEnabled(true);

		solveAction->setText(tr("Solve system"));
		solveAction->setEnabled(true);
		funEdi->setWindowTitle(tr("Function editor"));
		funEdiAction->setText(tr("Functions"));
		infoTextEdit->setEnabled(true);
		statusLabel->setText(tr("Edit: Functions")+"   "+tr("Grid:")+" "+tmp.setNum(sta.fgrids));

		Point->setEnabled(true);
		Triangle->setEnabled(true);
		Quad->setEnabled(true);

	}

	// Are we in implicite function mode?
	if(sta.funMod==IMPLICITE) {
		Function0->setText("I0");
		Function1->setText("I1");
		Function2->setText("I2");
		Function0->setChecked(iso[0].drawIso);
		Function1->setChecked(iso[1].drawIso);
		Function2->setChecked(iso[2].drawIso);
		Function0->setEnabled(true);
		Function1->setEnabled(true);
		Function2->setEnabled(true);

		ScaleBox->setToolTip(tr("Scale iso functions"));
		zPlus->setEnabled(true);
		zMinus->setEnabled(true);
		zscaleLabel->setEnabled(true);

		solveAction->setText(tr("For solving switch to function-mode"));
		solveAction->setEnabled(false);
		funEdi->setWindowTitle(tr("Iso editor"));
		funEdiAction->setText(tr("Iso functions"));
		infoTextEdit->setEnabled(false);
		statusLabel->setText(tr("Edit: Isosurfaces")+"   "+tr("Grid:")+" "+tmp.setNum(sta.tgrids));

		Point->setEnabled(true);
		Triangle->setEnabled(true);
		Quad->setEnabled(false);
	}

	// Or in parameter mode?
	if(sta.funMod==PARAMETER) {

		Function0->setEnabled(false);
		Function1->setEnabled(false);
		Function2->setEnabled(false);

		Function0->setChecked(iso[0].drawIso);
		Function1->setChecked(iso[1].drawIso);
		Function2->setChecked(iso[2].drawIso);

		ScaleBox->setToolTip(tr("Scale parametric system"));
		zPlus->setEnabled(false);
		zMinus->setEnabled(false);
		zscaleLabel->setEnabled(false);

		solveAction->setText(tr("For solving switch to function-mode"));
		solveAction->setEnabled(false);
		funEdi->setWindowTitle(tr("Parameter editor"));
		funEdiAction->setText(tr("Parameters"));
		infoTextEdit->setEnabled(false);
		statusLabel->setText(tr("Edit: Parameter")+"   "+tr("Grid:")+" "+tmp.setNum(sta.fgrids));

		Point->setEnabled(true);
		Triangle->setEnabled(true);
		Quad->setEnabled(true);
	}

	//Toggle labels on animation
	if(!ani.active) {
		if(sta.funMod==FUNCTION)
			infoTextEdit->setEnabled(true);
		runStopAction->setIcon(QIcon(":/images/play.png"));
		runStopAction->setToolTip(tr("Start animation"));
	}
	else {
		infoTextEdit->setEnabled(false);
		runStopAction->setIcon(QIcon(":/images/stop.png"));
		runStopAction->setToolTip(tr("Stop animation"));
	}
}


/* ********************************************************************** */
/* BENCHMARK                                                              */
/* ********************************************************************** */
void MaiWindow::benchmarkSlot() {

	const float roffset=0.15f;
	const int mSec=5000;
	long frames=0;

	char message[50];
	QString result(tr("Benchmark duration")+": "), tmp;
	QString currentTitle=this->windowTitle();

	// Backup and load benchmark-file
	saveToRam();
	if(!loadFile(MYSYSDIR+"benchmark.zhu")) {
		QMessageBox::critical(this, tr("Benchmark"), tr("Cannot open benchmark.zhu"));
		restoreFromRam();
		return;
	}

	// Initialize benchmark-viewer
	QApplication::setOverrideCursor(Qt::WaitCursor);
	aniEdi->animate((ani.active=false));
	oglWid->writeSettings();
	oglWid->move(QPoint(0,0));
	oglWid->resize(QSize(700,700));
	performOGLUpdate();
	oglWid->setWindowTitle(tr("Benchmarking now ..."));
	oglWid->raise();
	oglWid->activateWindow();

	// Perform benchmark
	benchReady=false;
	QTimer::singleShot(mSec, this, SLOT(benchTimerSlot()));
	while(!benchReady) {
			sta.rotX += roffset;
			sta.rotY += 3*roffset;
			sta.rotZ += 4*roffset;
			emit gotNewView();
			frames++;
			QApplication::processEvents();		// Give timer a chance!
	}
	QApplication::restoreOverrideCursor();

	// Prepare message-string
	tmp.setNum(mSec/1000.0, 'f', 1);
	result+=tmp+" sec.\nFPS = ";
	tmp.setNum(((int)frames)/(mSec/1000.0), 'f', 1);
	result+=tmp;

	// Show benchmark-info
	q2cstrcpy(message, result);
	QMessageBox::information(this,
		tr("Benchmark"),
		tr(message)
	);

	// Restore and perform updates
	restoreFromRam();
	aniEdi->animate(ani.active);
	morEdi->morph(mor.active);
	oglWid->setWindowTitle(tr("Viewer"));
	oglWid->readSettings();
	this->setWindowTitle(currentTitle);
	this->raise();
	this->activateWindow();

	// Update MaiWindow stuff
	updMaiWid();
	updMaiInf();

	// Reload textures
	for(int i=0; i<MAXFUN+1; i++)
		speEdi->loadTexture(texDir+tex[i].name, i);

	// Update other widgets
	funEdi->updFunWid();
	ligEdi->updLigWid();
	matEdi->updMatWid();
	entEdi->updEntWid();
	aniEdi->updAniWid();
	speEdi->updSpeWid();
	usrEdi->updUsrWid();

	// Dont forget multithreaded OpenGL lists
	iniITL();

	// Update OpenGL finally
	QApplication::setOverrideCursor(Qt::WaitCursor);
	PROC_EVENT;
	performOGLUpdate();
}


// Retranslate mainwidget
void MaiWindow::reTranslate() {

	// Menus
	fileMenu->setTitle(tr("File"));
	editorMenu->setTitle(tr("Editors"));
	settingsMenu->setTitle(tr("Settings"));
	scaleMenu->setTitle(tr("Scaling"));
	resetMenu->setTitle(tr("Reset"));
	windowsMenu->setTitle(tr("Windows"));
	helpMenu->setTitle(tr("Help"));

	// Menu entries
	newAction->setText(tr("&New"));
	newAction->setShortcut(tr("Ctrl+N"));
	openAction->setText(tr("&Open ..."));
	openAction->setShortcut(tr("Ctrl+O"));
	saveAsAction->setText(tr("&Save as"));
	saveAsAction->setShortcut(tr("Ctrl+S"));
	saveSolAction->setText(tr("Save solutions"));
	savePicAction->setText(tr("Save picture"));
	printAction->setText(tr("Print picture"));
	exitAction->setText(tr("&Exit"));
	exitAction->setShortcut(tr("Ctrl+E"));

	lightAction->setText(tr("Lights"));
	materialAction->setText(tr("Surfaces"));
	animateAction->setText(tr("Animation"));
	morphAction->setText(tr("Morphing"));
	entireAction->setText(tr("Global"));
	solEdiAction->setText(tr("Solver report"));
	usrEdiAction->setText(tr("User defined"));
	funEdiAction->setText(tr("Functions"));
	toFrontAction->setText(tr("Show open editors"));
	toBackAction->setText(tr("Hide open editors"));
	showAllAction->setText(tr("Open all editors"));

	pictureAction->setText(tr("Picture"));
	sfineAction->setText(tr("micro"));
	fineAction->setText(tr("fine"));
	normalAction->setText(tr("normal"));
	coarseAction->setText(tr("coarse"));
	resetViewAction->setText(tr("View only"));
	resetScaleAction->setText(tr("Scaling only"));
	resetLightAction->setText(tr("Lights only"));
	resetAllAction->setText(tr("All together"));
	fontAction->setText(tr("General"));
	speEdiAction->setText(tr("Special"));
	dirEdiAction->setText(tr("Directories"));

	systemAction->setText(tr("System"));
	benchmarkAction->setText(tr("Benchmark"));

	helpAction->setText(tr("Help"));
	demoAction->setText(tr("Demo"));
	helpAction->setShortcut(tr("F1"));
	aboutAction->setText(tr("About Zhu3D"));
	aboutQtAction->setText(tr("About Qt"));

	// Destroy toolbars first
	delete fileToolBar;
	delete zoomToolBar;
	delete transToolBar;
	delete extraToolBar;

	// Recrate them
	createToolBars();

	// Dynamic stuff; renew all instead of fiddling around
	updMaiWid();
	updMaiInf();
}
